/*
 * Program Name
 *  File2Msg - File to a Message
 *
 * Last date of modification
 *  Feb 15, 2010
 *
 * Description
 *  This program will read an entire file and write its contents as
 *  as a message to a queue. This program was written to handle
 *  data with embedded hex '00'.
 *
 *  This program can set the MQ_STR format field if the data is
 *  all character data.
 *
 *  The following is the logic in this program:
 *
 *  - Initialize and allocated memory.
 *  - Get the command line parameters.
 *  - Connect to the queue manager.
 *  - Open output queue
 *  - Calculate file size, allocate buffer and read file
 *  - put the message on the queue
 *  - Close output queue
 *  - Disconnect from the queue manager.
 *  - Terminate
 *
 * Copyright(C), 2001-2010, Roger Lacroix
 * Apache 2 License
 */

#ifdef MVS

#define  _XOPEN_SOURCE_EXTENDED 1

/*
 * MVS (OS/390) only pragma(s).
 */
#pragma runopts(ENV(MVS),PLIST(MVS),NOEXECOPS,\
NOTEST(NONE,INSPICD,PROMPT) )

#pragma csect(code,"F2M_MAIN")
#pragma csect(static,"F2M_WS")
#pragma comment( copyright , "Roger Lacroix." )
#endif

#pragma subtitle( "Load module eye-catchers" )
/*
 * Program build time eye-catchers for executable.
 */
static const char * EYECATCHER_EyeCatcher  = "File2Msg " ;
static const char * EYECATCHER_Version     = "1.2.1" ;
static const char * EYECATCHER_Datestamp   = __DATE__ ;
static const char * EYECATCHER_Timestamp   = __TIME__ ;
static const char * EYECATCHER_Filename    = __FILE__ ;

/*
 * Headers
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
   /* includes for MQI */
#include <cmqc.h>

#ifdef WIN32
#define  MAXPATH _MAX_PATH
#else
#define  MAXPATH PATH_MAX
#endif

/*
 * Global variables
 */
MQCHAR   QMgrName[MQ_Q_MGR_NAME_LENGTH];       // queue manager name
MQCHAR   QName[MQ_Q_NAME_LENGTH];              // queue name
MQCHAR   RQName[MQ_Q_NAME_LENGTH];             // reply-to-queue name
MQCHAR   RemoteQMgrName[MQ_Q_MGR_NAME_LENGTH]; // remote queue manager name

char    *pBuffer;
char     StringDataFlag='N';
char     FileName[MAXPATH];

char* usage = " -f filename -m qmgr -q queue [-s] [-r replyToQ] [-n remoteQMgrName]\n\
-f filename       The name of the input file               \n\
-m qmgr           The name of the queue manager            \n\
-q queue          The name of the output queue             \n\
-s                Flag to signal that data is all string [Optional]\n\
-r replytoq       The name of the reply to queue         [Optional]\n\
-n remoteQMgrName Remote queue manager name              [Optional]\n\
-?                Display this help message              [Optional]\n\
";

/*
 * Prototypes
 */
void GetParameters(int  argc, char *argv[]);
long FileSize (char  *pFileName);
int  ReadTheFile (char  *);

/*
 * Mainline
 */
int main(int argc, char **argv)
{
   MQOD     od = {MQOD_DEFAULT};    /* Object Descriptor             */
   MQMD     md = {MQMD_DEFAULT};    /* Message Descriptor            */
   MQPMO   pmo = {MQPMO_DEFAULT};   /* put message options           */

   MQHCONN  Hcon;                   /* connection handle             */
   MQHOBJ   Hobj;                   /* object handle                 */
   MQLONG   O_options;              /* MQOPEN options                */
   MQLONG   C_options;              /* MQCLOSE options               */
   MQLONG   CompCode;               /* completion code               */
   MQLONG   OpenCode;               /* MQOPEN completion code        */
   MQLONG   Reason;                 /* reason code                   */
   MQLONG   CReason;                /* reason code for MQCONN        */
   int      buffersize;
   int      BytesActuallyRead;

   // printf( "argc=%d\n", argc );

   if ( (argc < 7) || (argc > 12) )
   {
      printf( "Usage:\n%s%s\n", argv[0], usage );
      exit(-1);
   }

   printf("Starting File2Msg\n");

   /*
    * Clear all queue and qmgr names.
    */
   memset( QMgrName, ' ', MQ_Q_MGR_NAME_LENGTH );
   memset( QName,  ' ', MQ_Q_NAME_LENGTH );
   memset( RQName, ' ', MQ_Q_NAME_LENGTH );
   memset( RemoteQMgrName, ' ', MQ_Q_MGR_NAME_LENGTH );

   GetParameters(argc, argv);

   buffersize = FileSize( FileName );

   if (buffersize > 100 * 1024 * 1024)
   {
      printf("File size of %ld is too large to be an MQ message\n", buffersize);
      exit( -1 );
   }

   /* Connect to queue manager */
   MQCONN(QMgrName,                  /* queue manager                  */
          &Hcon,                   /* connection handle              */
          &CompCode,               /* completion code                */
          &CReason);               /* reason code                    */


   printf("MQCONN ended with cc %ld rc %ld\n", CompCode, CReason);
   /* report reason and stop if it failed     */
   if (CompCode == MQCC_FAILED)
   {
     exit( (int)CReason );
   }

   /*   Use parameter as the name of the target queue  */
   strncpy(od.ObjectName, QName, (size_t)MQ_Q_NAME_LENGTH);

   strncpy(od.ObjectQMgrName, RemoteQMgrName, (size_t)MQ_Q_MGR_NAME_LENGTH);

   printf("Target queue is %s\n", od.ObjectName);

   /* Open the target message queue for output */
   O_options = MQOO_OUTPUT           /* open queue for output        */
               + MQOO_FAIL_IF_QUIESCING;

   MQOPEN(Hcon,                      /* connection handle            */
          &od,                       /* object descriptor for queue  */
          O_options,                 /* open options                 */
          &Hobj,                     /* object handle                */
          &OpenCode,                 /* MQOPEN completion code       */
          &Reason);                  /* reason code                  */

   /* report reason, if any; stop if failed      */
   printf("MQOPEN ended with reason code %ld\n", Reason);

   if (OpenCode == MQCC_OK)
   {
      CompCode = OpenCode;        /* use MQOPEN result for initial test */

      if (buffersize != 0)
      {
        printf("File size is %ld\n",buffersize);
        pBuffer = malloc( buffersize);
        BytesActuallyRead = ReadTheFile( FileName );
      }

      /* Put each buffer to the message queue */
      if (BytesActuallyRead > 0)
      {
         /* Did they specify a reply to queue name ? */
         if ( RQName[0] == ' ' )
         {
            md.MsgType = MQMT_DATAGRAM;
            memset(md.ReplyToQ, ' ', MQ_Q_NAME_LENGTH);
         }
         else
         {
            md.MsgType = MQMT_REQUEST;
            strncpy(md.ReplyToQ, RQName, MQ_Q_NAME_LENGTH);
         }

         memcpy(md.CorrelId, MQCI_NONE, sizeof(md.CorrelId));
         memcpy(md.MsgId, MQMI_NONE, sizeof(md.MsgId));

         if ( StringDataFlag == 'Y' )
         {
            memcpy(md.Format,
                    MQFMT_STRING, (size_t)MQ_FORMAT_LENGTH);
            printf("md.Format=%s\n", MQFMT_STRING);
         }

         MQPUT(Hcon,                /* connection handle               */
               Hobj,                /* object handle                   */
               &md,                 /* message descriptor              */
               &pmo,                /* default options (datagram)      */
               BytesActuallyRead,   /* buffer length                   */
               pBuffer,             /* message buffer                  */
               &CompCode,           /* completion code                 */
               &Reason);            /* reason code                     */

         /* report reason, if any */
         printf("MQPUT ended with reason code %ld\n", Reason);
      }
      else
      {
         /* satisfy end condition when empty line is read */
         CompCode = MQCC_FAILED;
      }

      if ( pBuffer != NULL )
      {
         free(pBuffer);
      }

      /* Close the target queue (if it was opened)  */
      if (OpenCode != MQCC_FAILED)
      {
         C_options = 0;                  /* no close options             */
         MQCLOSE(Hcon,                   /* connection handle            */
                 &Hobj,                   /* object handle               */
                 C_options,
                 &CompCode,               /* completion code             */
                 &Reason);                /* reason code                 */

         printf("MQCLOSE ended with reason code %ld\n", Reason);
      }
   }
   else
   {
     printf("unable to open queue for output\n");
   }

   /* Disconnect from MQM if not already connected  */
   if (CReason != MQRC_ALREADY_CONNECTED)
   {
      MQDISC(&Hcon,                   /* connection handle            */
             &CompCode,               /* completion code              */
             &Reason);                /* reason code                  */

      printf("MQDISC ended with reason code %ld\n", Reason);
   }

   printf("File2Msg is ending.\n");
   return(0);
 }

void GetParameters(int argc, char *argv[])
{
   #define MaxParams 8

   char *ParmkeyUC[MaxParams] = { "-F", "-M", "-Q", "-R", "-S", "-N", ",", "-?"};
   char *ParmkeyLC[MaxParams] = { "-f", "-m", "-q", "-r", "-s", "-n", ",", "-?"};
   int i,j;
   int rc=0;

   /* Prior to processing the parameters, establish the default */
   /* values this is done before calling GetParameters          */

   if (argc >1)
   {
      for (i=1; i<argc; i++)
      {
         for (j=0; j<MaxParams; j++)
         {
            if ( (!strncmp( argv[i], ParmkeyUC[j], 2)) ||
                 (!strncmp( argv[i], ParmkeyLC[j], 2)) )
            {
                break; /* fall out to switch stmt */
            }
         };

         switch (j)
         {
            case 0:                      /* -f */
               if (i+1 < argc)
               {
                   strncpy( FileName, argv[i+1], MAXPATH );
                   i++;
               }
               else
               {
                   rc=9;
               }
               break;

            case 1:                      /* -m */
               if (i+1 < argc)
               {
                   strncpy(QMgrName, argv[i+1], MQ_Q_MGR_NAME_LENGTH);
                   i++;
               }
               else
               {
                   rc=9;
               }
               break;

            case 2:                      /* -q */
               if (i+1 < argc)
               {
                   strncpy(QName, argv[i+1], MQ_Q_NAME_LENGTH);
                   i++;
               }
               else
               {
                   rc=9;
               }
               break;

            case 3:                      /* -r */
               if (i+1 < argc)
               {
                   strncpy(RQName, argv[i+1], MQ_Q_NAME_LENGTH);
                   i++;
               }
               else
               {
                   rc=9;
               }
               break;

            case 4:                      /* -s */
               StringDataFlag = 'Y';
               break;

            case 5:                      /* -n */
              if (i+1 < argc)
              {
                  strncpy(RemoteQMgrName, argv[i+1], MQ_Q_MGR_NAME_LENGTH);
                  i++;
              }
              else
              {
                  rc=9;
              }
              break;

           case 6:                      /* , */
              // ignore the comma
              break;

           case 7:                      /* -? */
              printf( "Usage:\n%s%s\n", argv[0], usage );
              exit(-1);
              break;

           default:/* parameter error */
              printf( "Error with keyword parameter!\n" );
              printf( "Parameter=%s\n", argv[i] );
              exit(-1);
              break;
              /* ignore it */
         };
      };
   }

   if ( *(QName) == ' ' || *(QMgrName) == ' ' || *(FileName) == ' ' )
   {
      printf( "Usage:\n%s%s\n", argv[0], usage );
      exit(-1);
   }

   if ( rc==9 )
   {
      printf( "Usage:\n%s%s\n", argv[0], usage );
      exit(-1);
   }

   return;
}

/*
 * FILE_SIZE returns the length of the file given as its argument.
 * The file is required to be opened already. Note that the position
 * witin the file is preserved through the call. This function will
 * only work properly with fixed format files.
 */

long FileSize (char  *pFileName)
{
   auto long int size = 0L; /* Size of the file. */
   FILE   *stream;

   if( (stream = fopen( pFileName, "r" )) == NULL )
   {
     printf("Error: File %s does not exist!!!!\n", pFileName );
   }
   else
   {
      fseek (stream, 0L, SEEK_END);
      size = ftell (stream);
//    size = ftell (stream) + 1; /* Offsets run 0 to (size - 1). */

      if (size == 0)
      {
        printf("Error: File %s is empty!!!!\n", pFileName );
      }

      fclose( stream );
   }

   return (size);
}

/*
 * Read the file
 */
int  ReadTheFile(char  *pFileName)
{
   FILE   *stream;
   int    BytesRead=0;
   int    ch;
   int    rc;
   char   *ptemp;

   ptemp = pBuffer;

   /* Open file to read line from: */
   if ( (stream = fopen( pFileName, "rb" )) == NULL )
   {
      rc = -1;
   }
   else
   {
      ch = fgetc( stream );
      while ( !feof( stream ) )
      {
         *ptemp++ = (char)ch;
         BytesRead++;
         ch = fgetc( stream );
      }

      fclose( stream );

      printf("Read %d bytes from file %s\n", BytesRead, pFileName );

      rc = BytesRead;
   }
   return rc;
}
